@@ -30,12 +30,14 @@
  *  
  *  $Author: peter $
  *
- *  $Id: netflow_v9.c 70 2006-05-17 08:38:01Z peter $
+ *  $Id: netflow_v9.c 97 2008-02-21 09:50:02Z peter $
  *
- *  $LastChangedRevision: 70 $
+ *  $LastChangedRevision: 97 $
  *	
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -47,8 +49,6 @@
 #include <time.h>
 #include <netinet/in.h>
 
-#include "config.h"
-
 #ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif
@@ -82,6 +82,7 @@ typedef struct input_translation_s {
 	uint32_t	zero_index;
 	uint32_t    packet_offset;
 	uint32_t    byte_offset;
+	uint32_t	ICMP_offset;
 	translation_element_t element[NumElements];
 } input_translation_t;
 
@@ -114,11 +115,11 @@ static struct element_info_s {
 	{ 2, 2 },	//  7 - NF9_L4_SRC_PORT
 	{ 4, 4 },	//  8 - NF9_IPV4_SRC_ADDR
 	{ 2, 2 },	//  9 - NF9_SRC_MASK
-	{ 2, 2 },	// 10 - NF9_INPUT_SNMP
+	{ 2, 4 },	// 10 - NF9_INPUT_SNMP
 	{ 2, 2 },	// 11 - NF9_L4_DST_PORT
 	{ 4, 4 },	// 12 - NF9_IPV4_DST_ADDR
 	{ 2, 2 },	// 13 - NF9_DST_MASK
-	{ 2, 2 },	// 14 - NF9_OUTPUT_SNMP
+	{ 2, 4 },	// 14 - NF9_OUTPUT_SNMP
 	{ 4, 4 },	// 15 - NF9_IPV4_NEXT_HOP
 	{ 2, 2 },	// 16 - NF9_SRC_AS
 	{ 2, 2 },	// 17 - NF9_DST_AS
@@ -137,7 +138,7 @@ static struct element_info_s {
 	{ 4, 4 },	// 29 - NF9_IPV6_SRC_MASK
 	{ 4, 4 },	// 30 - NF9_IPV6_DST_MASK
 	{ 4, 4 },	// 31 - NF9_IPV6_FLOW_LABEL
-	{ 4, 4 },	// 32 - NF9_ICMP_TYPE
+	{ 2, 2 },	// 32 - NF9_ICMP_TYPE
 
 	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 			// 33 - 37 not implemented
 
@@ -327,15 +328,22 @@ uint32_t	zero_index  = table->zero_index;
 
 	output_length = CheckElementLength(element, input_template[element].length);
 	if ( output_length ) { 
-	/*
+/*
 		printf("Index: %u Elem %i, IO %u, OO %u, len: %u\n", 
 			input_index, element, input_template[element].offset, *offset, input_template[element].length);
-	*/
+*/
 		table->element[input_index].output_offset 	= *offset;
 		table->element[input_index].input_offset 	= input_template[element].offset;
 		table->element[input_index].length 			= input_template[element].length;
+		if ( ( element == NF9_INPUT_SNMP || element == NF9_OUTPUT_SNMP ) && output_length == 4 )  {
+			table->element[input_index].input_offset 	+= 2;
+			table->element[input_index].length 			= 2;
+
+			(*offset)	+= 2;
+		} else {
+			(*offset)	+= output_length;
+		}
 		table->input_index++;
-		(*offset)	+= output_length;
 	} else {
 	/*
 		printf("Zero: %u, Elem: %i,  OO %u, len: %u\n", 
@@ -367,8 +375,9 @@ uint32_t			offset;
 
 	// clear current table
 	memset((void *)table->element, 0, NumElements * sizeof(translation_element_t));
-	table->updated  = time(NULL);
-	table->flags	= 0;
+	table->updated  	= time(NULL);
+	table->flags		= 0;
+	table->ICMP_offset	= 0;
 
 	// printf("[%u] Fill translation table %u\n", exporter->exporter_id, id);
 
@@ -445,6 +454,15 @@ uint32_t			offset;
 	table->input_record_size  = input_record_size;
 	table->output_record_size = offset;
 
+	/* ICMP hack for v9, until file version 2 for FNF is fully impemented */
+	if ( input_template[NF9_ICMP_TYPE].offset != 0 ) {
+		if ( input_template[NF9_ICMP_TYPE].length == 2 ) 
+			table->ICMP_offset = input_template[NF9_ICMP_TYPE].offset;
+		else
+			syslog(LOG_ERR, "Process_v9: Unexpected ICMP type field length: %d \n", 
+				input_template[NF9_ICMP_TYPE].length);
+	}
+
 	/*
 	printf("Table %u Flags: %u, index: %u, Zero: %u input_size: %u, output_size: %u\n", 
 		table->id, table->flags, table->input_index, table->zero_index, table->input_record_size, table->output_record_size);
@@ -551,7 +569,7 @@ char				*string;
 		}
 
 		// check for enough space in output buffer
-		if ( (data_header->size + table->output_record_size) > OUTPUT_BUFF_SIZE ) {
+		if ( (data_header->size + table->output_record_size) > BUFFSIZE ) {
 			// this should really never occur, because the buffer gets flushed ealier
 			syslog(LOG_ERR,"Process_v9: output buffer size error. Abort v9 record processing");
 			return writeto;
@@ -663,6 +681,12 @@ char				*string;
 			}
 		} // End for
 
+		// Ungly ICMP hack for v9, until file version 2 for FNF is implemented 
+		if ( ( data_record->prot == IPPROTO_ICMP || data_record->prot == IPPROTO_ICMPV6 ) && 
+				table->ICMP_offset ) { // it's an ICMP record
+			data_record->dstport = ntohs(Get_val16((void *)&in[table->ICMP_offset]));
+		}
+
 		First = data_record->first;
 		Last  = data_record->last;
 
@@ -746,7 +770,7 @@ char				*string;
 		if ( verbose ) {
 			master_record_t master_record;
 			ExpandRecord((common_record_t *)data_record, &master_record);
-		 	format_file_block_record(&master_record, 1, &string, 0);
+		 	format_file_block_record(&master_record, 1, &string, 0, 0);
 			printf("%s\n", string);
 		}
 
@@ -759,7 +783,7 @@ char				*string;
 
 		// buffer size sanity check
 		bsize = (pointer_addr_t)data_record - (pointer_addr_t)data_header;
-		if ( bsize >= OUTPUT_BUFF_SIZE ) {
+		if ( bsize >= BUFFSIZE ) {
 			syslog(LOG_ERR,"Process v9: Output buffer overflow! Flush buffer and skip records.");
 			return (void *)data_record;
 		}
@@ -782,7 +806,7 @@ ssize_t				size_left;
 
 	size_left = in_buff_cnt;
 	if ( size_left < NETFLOW_V9_HEADER_LENGTH ) {
-		syslog(LOG_ERR, "Process_v9: Too little data for v9 packets: '%u'\n", size_left);
+		syslog(LOG_ERR, "Process_v9: Too little data for v9 packets: '%lli'\n", (long long)size_left);
 		return writeto;
 	}
 
@@ -859,7 +883,8 @@ ssize_t				size_left;
 		}
 
 		if ( flowset_length > size_left ) {
-			syslog(LOG_ERR,"Process_v9: flowset length error. Expected bytes: %u but buffersize: %u\n", flowset_length, size_left);
+			syslog(LOG_ERR,"Process_v9: flowset length error. Expected bytes: %u but buffersize: %lli\n", 
+				flowset_length, (long long)size_left);
 			size_left = 0;
 			continue;
 		}
@@ -1001,6 +1026,11 @@ uint32_t	count, record_length;
 	record_length 				+= element_info[NF9_DST_AS].min;
 	count++;
 
+	fields->record[count].type   = htons(NF9_ICMP_TYPE);
+	fields->record[count].length = htons(element_info[NF9_ICMP_TYPE].min);
+	record_length 				+= element_info[NF9_ICMP_TYPE].min;
+	count++;
+
 	if ( (flags & FLAG_IPV6_ADDR) != 0 ) { // IPv6 addresses
 		fields->record[count].type   = htons(NF9_IPV6_SRC_ADDR);
 		fields->record[count].length = htons(element_info[NF9_IPV6_SRC_ADDR].min);
@@ -1059,12 +1089,13 @@ static uint64_t	boot_time;	// in msec
 static uint32_t	last_flags, common_block_size;
 static int	record_count, template_count, flowset_count, packet_count;
 uint32_t	required_size, t1, t2;
+uint16_t	icmp;
 void		*endwrite;
 time_t		now = time(NULL);
 
 /*
 	char		*string;
-	format_file_block_record(master_record, 1, &string, 0);
+	format_file_block_record(master_record, 1, &string, 0, 0);
 	printf("%s\n", string);
 */
 	if ( !v9_output_header->unix_secs ) {	// first time a record is added
@@ -1210,9 +1241,19 @@ time_t		now = time(NULL);
   	master_record->srcas	= htons(master_record->srcas);
   	master_record->dstas	= htons(master_record->dstas);
 
+	if ( master_record->prot == IPPROTO_ICMP || master_record->prot == IPPROTO_ICMPV6  ) { // it's an ICMP
+		icmp = master_record->dstport;
+		master_record->dstport = 0;
+	} else {
+		icmp = 0;
+	}
+
 	memcpy(peer->writeto, (void *)&master_record->first,common_block_size);
 	peer->writeto = (void *)((pointer_addr_t)peer->writeto + common_block_size);
 
+	memcpy(peer->writeto, (void *)&icmp,2);
+	peer->writeto = (void *)((pointer_addr_t)peer->writeto + 2);
+	
 	if ((master_record->flags & FLAG_IPV6_ADDR) != 0 ) { // IPv6
 		master_record->v6.srcaddr[0] = htonll(master_record->v6.srcaddr[0]);
 		master_record->v6.srcaddr[1] = htonll(master_record->v6.srcaddr[1]);
