@@ -32,16 +32,19 @@
  *  
  *  $Author: peter $
  *
- *  $Id: nfdump.c 76 2006-05-21 16:29:48Z peter $
+ *  $Id: nfdump.c 98 2008-02-22 09:13:12Z peter $
  *
- *  $LastChangedRevision: 76 $
+ *  $LastChangedRevision: 98 $
  *	
  *
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <errno.h>
 #include <time.h>
 #include <string.h>
@@ -52,8 +55,6 @@
 #include <fcntl.h>
 #include <sys/resource.h>
 
-#include "config.h"
-
 #ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif
@@ -62,32 +63,33 @@
 #include "nfnet.h"
 #include "nf_common.h"
 #include "netflow_v5_v7.h"
+#include "rbtree.h"
 #include "nftree.h"
 #include "nfprof.h"
 #include "nfdump.h"
 #include "nfstat.h"
+#include "ipconv.h"
 #include "version.h"
 #include "launch.h"
 #include "util.h"
+#include "flist.h"
 #include "panonymizer.h"
 
 /* hash parameters */
 #define HashBits 20
 #define NumPrealloc 128000
 
-#define AGGR_SIZE 5
+#define AGGR_SIZE 7
 
 
 /* Global Variables */
 FilterEngine_data_t	*Engine;
-uint32_t			byte_limit, packet_limit;
-int 				byte_mode, packet_mode;
 
 /* Local Variables */
-static char const *rcsid 		  = "$Id: nfdump.c 76 2006-05-21 16:29:48Z peter $";
+static char const *rcsid 		  = "$Id: nfdump.c 98 2008-02-22 09:13:12Z peter $";
 static uint64_t total_bytes;
 static uint32_t total_flows;
-static uint32_t skipped_flows;
+static uint32_t skipped_records;
 static time_t t_first_flow, t_last_flow;
 
 extern const uint16_t MAGIC;
@@ -149,7 +151,7 @@ extern const uint16_t VERSION;
  *    { "formatname", format_special, FORMAT_definition, NULL },
  *   The first parameter is the name of your format as recognized on the command line as -o <formatname>
  *   The second parameter is always 'format_special' - the printing function.
- *   The third parameter is you format definition as defined in #define.
+ *   The third parameter is your format definition as defined in #define.
  *   The forth parameter is always NULL for user defined formats.
  * 5. Recompile nfdump
  */
@@ -181,6 +183,9 @@ struct printmap_s {
 // all records should be version 5
 #define FLOW_VERSION 5
 
+/* exported fuctions */
+void LogError(char *format, ...);
+
 /* Function Prototypes */
 static void usage(char *name);
 
@@ -188,11 +193,11 @@ static int ParseAggregateMask( char *arg, uint64_t *AggregateMasks, uint16_t *Ag
 
 static int ParseCryptoPAnKey ( char *s, char *key );
 
-static void PrintSummary(stat_record_t *stat_record);
+static void PrintSummary(stat_record_t *stat_record, int plain_numbers);
 
 static stat_record_t process_data(char *wfile, int element_stat, int flow_stat, int sort_flows,
 	printer_t print_header, printer_t print_record, time_t twin_start, time_t twin_end, 
-	uint64_t limitflows, uint64_t *AggregateMasks, int anon, int zero_flows);
+	uint64_t limitflows, uint64_t *AggregateMasks, int anon, int tag, int compress);
 
 /* Functions */
 static void usage(char *name) {
@@ -202,17 +207,20 @@ static void usage(char *name) {
 					"-a\t\tAggregate netflow data.\n"
 					"-A <expr>[/net]\tHow to aggregate: ',' sep list of 'srcip dstip srcport dstport'\n"
 					"\t\tor subnet aggregation: srcip4/24, srcip6/64.\n"
-					"-r\t\tread input from file\n"
-					"-w\t\twrite output to file\n"
+					"-r <file>\tread input from file\n"
+					"-w <file>\twrite output to file\n"
 					"-f\t\tread netflow filter from file\n"
 					"-n\t\tDefine number of top N. \n"
 					"-c\t\tLimit number of records to display\n"
+					"-D <dns>\tUse nameserver <dns> for host lookup.\n"
 					"-S\t\tGenerate netflow statistics info.\n"
 					"-s <expr>[/<order>]\tGenerate statistics for <expr>: \n"
+					"-N\t\tPrint plain numbers in summary line\n"
 					"\t\tsrcip, dstip, ip, srcport, dstport, port, srcas, dstas, as, inif, outif, proto\n"
 					"\t\tand ordered by <order>: packets, bytes, flows, bps pps and bpp.\n"
 					"-q\t\tQuiet: Do not print the header and bottom stat lines.\n"
-					"-z\t\tZero flows - dumpfile contains only statistics record.\n"
+					"-j <file>\tCompress/Uncompress file.\n"
+					"-z\t\tCompress flows in output file. Used in combination with -w.\n"
 					"-l <expr>\tSet limit on packets for line and packed output format.\n"
 					"-K <key>\tAnonymize IP addressses using CryptoPAn with key <key>.\n"
 					"\t\tkey: 32 character string or 64 digit hex string starting with 0x.\n"
@@ -220,7 +228,7 @@ static void usage(char *name) {
 					"-M <expr>\tRead input from multiple directories.\n"
 					"-I \t\tPrint netflow summary statistics info from file, specified by -r.\n"
 					"\t\t/dir/dir1:dir2:dir3 Read the same files from '/dir/dir1' '/dir/dir2' and '/dir/dir3'.\n"
-					"\t\treqquests either -r filename or -R firstfile:lastfile without pathnames\n"
+					"\t\trequests either -r filename or -R firstfile:lastfile without pathnames\n"
 					"-m\t\tPrint netflow data date sorted. Only useful with -M\n"
 					"-R <expr>\tRead input from sequence of files.\n"
 					"\t\t/any/dir  Read all files in that directory.\n"
@@ -233,18 +241,37 @@ static void usage(char *name) {
 					"\t\t extended Even more information.\n"
 					"\t\t pipe     '|' separated, machine parseable output format.\n"
 					"\t\t\tmode may be extended by '6' for full IPv6 listing. e.g.long6, extended6.\n"
+					"-v <file>\tverify netflow data file. Print version and blocks.\n"
 					"-X\t\tDump Filtertable and exit (debug option).\n"
 					"-Z\t\tCheck filter syntax and exit.\n"
 					"-t <time>\ttime window for filtering packets\n"
 					"\t\tyyyy/MM/dd.hh:mm:ss[-yyyy/MM/dd.hh:mm:ss]\n", name);
 } /* usage */
 
+/* 
+ * some modules are needed for daemon code as well as normal stdio code 
+ * therefore a generic LogError is defined, which maps in this case
+ * to stderr
+ */
+void LogError(char *format, ...) {
+va_list var_args;
+
+	va_start(var_args, format);
+	vfprintf(stderr, format, var_args);
+	va_end(var_args);
+
+} // End of LogError
+
+
 static int ParseAggregateMask( char *arg, uint64_t *AggregateMasks, uint16_t *Aggregate_bits ) {
 char *p, *q;
 uint32_t	subnet;
 
-	// preset Aggregate Mask array
+	// preset Aggregate Mask array, for those fields not affected
 	AggregateMasks[0] = ~( MaskSrcPort | MaskDstPort );
+	AggregateMasks[5] = ~( MaskSrcAS | MaskDstAS );
+	AggregateMasks[5] = ~( MaskSrcAS | MaskDstAS );
+	AggregateMasks[6] = ~MaskProto;
 	*Aggregate_bits 	  = 0;
 
 	subnet = 0;
@@ -327,6 +354,15 @@ uint32_t	subnet;
 		} else if ( strcasecmp(p, "dstport" ) == 0 ) {
 			AggregateMasks[0] |= MaskDstPort;
 			*Aggregate_bits |= Aggregate_DSTPORT;
+		} else if (  strcasecmp(p, "srcas" ) == 0 ) {
+			AggregateMasks[5] |= MaskSrcAS;
+			*Aggregate_bits |= Aggregate_SRCAS;
+		} else if ( strcasecmp(p, "dstas" ) == 0 ) {
+			AggregateMasks[5] |= MaskDstAS;
+			*Aggregate_bits |= Aggregate_DSTAS;
+		} else if ( strcasecmp(p, "proto" ) == 0 ) {
+			AggregateMasks[6] |= MaskProto;
+			*Aggregate_bits |= Aggregate_PROTO;
 
 
 		} else {
@@ -335,6 +371,14 @@ uint32_t	subnet;
 		}
 		p = strtok(NULL, ",");
 	}
+
+	// if we aggregate srcip and dstip, we retain the AS info and do not mask them out
+	// as this info may be important
+	if ( (*Aggregate_bits & ( Aggregate_SRCIP + Aggregate_DSTIP )) == ( Aggregate_SRCIP + Aggregate_DSTIP ) ) {
+		AggregateMasks[5] = MaskSrcAS | MaskDstAS;
+		*Aggregate_bits |= Aggregate_SRCAS;
+		*Aggregate_bits |= Aggregate_DSTAS;
+	}
 	return 1;
 
 } /* End of ParseAggregateMask */
@@ -368,38 +412,43 @@ char numstr[3];
 
 } // End of ParseCryptoPAnKey
 
-static void PrintSummary(stat_record_t *stat_record) {
+static void PrintSummary(stat_record_t *stat_record, int plain_numbers) {
 static double	duration;
 uint64_t	bps, pps, bpp;
-char        byte_str[32], packet_str[32], bps_str[32], pps_str[32], bpp_str[32];
+char 		byte_str[32], packet_str[32], bps_str[32], pps_str[32], bpp_str[32];
 
 	bps = pps = bpp = 0;
 	duration = stat_record->last_seen - stat_record->first_seen;
 	duration += ((double)stat_record->msec_last - (double)stat_record->msec_first) / 1000.0;
 	if ( duration > 0 && stat_record->last_seen > 0 ) {
 		bps = ( stat_record->numbytes << 3 ) / duration;	// bits per second. ( >> 3 ) -> * 8 to convert octets into bits
-		pps = stat_record->numpackets / duration;		   // packets per second
-		bpp = stat_record->numbytes / stat_record->numpackets;  // Bytes per Packet
+		pps = stat_record->numpackets / duration;			// packets per second
+		bpp = stat_record->numbytes / stat_record->numpackets;	// Bytes per Packet
+	}
+	if ( plain_numbers ) {
+		printf("Summary: total flows: %llu, total bytes: %llu, total packets: %llu, avg bps: %llu, avg pps: %llu, avg bpp: %llu\n",
+			stat_record->numflows, stat_record->numbytes, stat_record->numpackets, bps, pps, bpp );
+	} else {
+		format_number(stat_record->numbytes, byte_str, VAR_LENGTH);
+		format_number(stat_record->numpackets, packet_str, VAR_LENGTH);
+		format_number(bps, bps_str, VAR_LENGTH);
+		format_number(pps, pps_str, VAR_LENGTH);
+		format_number(bpp, bpp_str, VAR_LENGTH);
+		printf("Summary: total flows: %llu, total bytes: %s, total packets: %s, avg bps: %s, avg pps: %s, avg bpp: %s\n",
+		(unsigned long long)stat_record->numflows, byte_str, packet_str, bps_str, pps_str, bpp_str );
 	}
-	format_number(stat_record->numbytes, byte_str, VAR_LENGTH);
-	format_number(stat_record->numpackets, packet_str, VAR_LENGTH);
-	format_number(bps, bps_str, VAR_LENGTH);
-	format_number(pps, pps_str, VAR_LENGTH);
-	format_number(bpp, bpp_str, VAR_LENGTH);
-	printf("Summary: total flows: %llu, total bytes: %s, total packets: %s, avg bps: %s, avg pps: %s, avg bpp: %s\n",
-		stat_record->numflows, byte_str, packet_str, bps_str, pps_str, bpp_str );
 
 } // End of PrintSummary
 
 stat_record_t process_data(char *wfile, int element_stat, int flow_stat, int sort_flows,
 	printer_t print_header, printer_t print_record, time_t twin_start, time_t twin_end, 
-	uint64_t limitflows, uint64_t *AggregateMasks, int anon, int zero_flows) {
-data_block_header_t in_flow_header, *out_flow_header;					
+	uint64_t limitflows, uint64_t *AggregateMasks, int anon, int tag, int compress) {
+data_block_header_t in_block_header, *out_block_header;					
 common_record_t 	*flow_record, *in_buff, *out_buff;
 master_record_t		master_record;
 void				*writeto;
 stat_record_t 		stat_record;
-uint32_t	NumReadRecords, file_blocks, buffer_size;
+uint32_t	file_blocks;
 int 		i, rfd, wfd, done, ret, do_stat, write_file, has_aggregate_mask;
 char 		*string;
 
@@ -407,7 +456,7 @@ char 		*string;
 extern int	Format14;
 #endif
 
-	out_flow_header = NULL;
+	out_block_header = NULL;
 	out_buff 		= NULL;
 	writeto			= NULL;
 
@@ -451,7 +500,7 @@ extern int	Format14;
 #endif
 
 	if ( wfile ) {
-		wfd = strcmp(wfile, "-") == 0 ? STDOUT_FILENO : OpenNewFile(wfile, &string);
+		wfd = strcmp(wfile, "-") == 0 ? STDOUT_FILENO : OpenNewFile(wfile, &string, compress);
 		if ( strcmp(wfile, "-") == 0 ) { // output to stdout
 			file_header_t	*file_header;
 			size_t			len;
@@ -467,7 +516,7 @@ extern int	Format14;
 			write(STDOUT_FILENO, (void *)file_header, len) ;
 
 		} else {
-			wfd = OpenNewFile(wfile, &string);
+			wfd = OpenNewFile(wfile, &string, compress);
 
 		}
 		if ( wfd < 0 ) {
@@ -477,30 +526,29 @@ extern int	Format14;
 				close(rfd);
 			return stat_record;
 		}
-		out_buff = malloc(OUTPUT_BUFF_SIZE);
+		out_buff = malloc(BUFFSIZE);
 		if ( !out_buff ) {
 			fprintf(stderr, "Buffer allocation error: %s", strerror(errno));
 			return stat_record;
 		}
-		out_flow_header 			= (data_block_header_t *)out_buff;
-		out_flow_header->size 		= 0;
-		out_flow_header->NumBlocks 	= 0;
-		out_flow_header->id			= DATA_BLOCK_TYPE_1;
-		out_flow_header->pad		= 0;
-		writeto = (void *)((pointer_addr_t)out_flow_header + sizeof(data_block_header_t));
+		out_block_header 			= (data_block_header_t *)out_buff;
+		out_block_header->size 		= 0;
+		out_block_header->NumBlocks	= 0;
+		out_block_header->id		= DATA_BLOCK_TYPE_1;
+		out_block_header->pad		= 0;
+		writeto = (void *)((pointer_addr_t)out_block_header + sizeof(data_block_header_t));
 	} else 
 		wfd = 0;
 
 	// allocate buffer suitable for netflow version
-	buffer_size = BUFFSIZE;
-	in_buff = (common_record_t *) malloc(buffer_size);
+	in_buff = (common_record_t *) malloc(BUFFSIZE);
 
 	if ( !in_buff ) {
 		perror("Memory allocation error");
 		close(rfd);
 		if ( write_file ) {
 			/* Write stat info and close file */
-			CloseUpdateFile(wfd, &stat_record, file_blocks, GetIdent(), &string );
+			CloseUpdateFile(wfd, &stat_record, file_blocks, GetIdent(), compress, &string );
 			if ( string != NULL )
 				fprintf(stderr, "%s\n", string);
 		} 
@@ -514,122 +562,42 @@ extern int	Format14;
 	done = 0;
 	while ( !done ) {
 
-#ifdef COMPAT14
-		if ( Format14 ) 
-			ret = Compat14_ReadHeader(rfd, &in_flow_header);
-		else
-			ret = read(rfd, &in_flow_header, sizeof(data_block_header_t));
-#else
-		ret = read(rfd, &in_flow_header, sizeof(data_block_header_t));
-#endif
+		// get next data block from file
+		ret = ReadBlock(rfd, &in_block_header, (void *)in_buff, &string);
 
-		if ( ret == 0 ) {
-			// EOF of rfd
-			rfd = GetNextFile(rfd, twin_start, twin_end, NULL);
-			if ( rfd < 0 ) {
-				if ( rfd == FILE_ERROR )
-					fprintf(stderr, "Can't read from file '%s': %s\n",GetCurrentFilename(), strerror(errno) );
-				done = 1;
-			} 
-			continue;
-		} else if ( ret == -1 ) {
-			fprintf(stderr, "Can't read from file '%s': %s\n",GetCurrentFilename(), strerror(errno) );
-			break;
-		}
-		total_bytes += ret;
+		switch (ret) {
+			case NF_CORRUPT:
+			case NF_ERROR:
+				if ( ret == NF_CORRUPT ) 
+					fprintf(stderr, "Skip corrupt data file '%s': '%s'\n",GetCurrentFilename(), string);
+				else 
+					fprintf(stderr, "Read error in file '%s': %s\n",GetCurrentFilename(), strerror(errno) );
+				// fall through - get next file in chain
+			case NF_EOF:
+				rfd = GetNextFile(rfd, twin_start, twin_end, NULL);
+				if ( rfd < 0 ) {
+					if ( rfd == NF_ERROR )
+						fprintf(stderr, "Read error in file '%s': %s\n",GetCurrentFilename(), strerror(errno) );
 
-		if ( in_flow_header.id != DATA_BLOCK_TYPE_1 ) {
-			fprintf(stderr, "Can't process block type %u\n", in_flow_header.id);
-			skipped_flows++;
-			continue;
-		}
-
-		NumReadRecords = in_flow_header.NumBlocks;
-
-		if ( in_flow_header.size > buffer_size ) {
-			void *tmp;
-			// Actually, this should never happen, but catch it anyway
-			if ( in_flow_header.size > MAX_BUFFER_SIZE ) {
-				// this is most likely corrupt
-				fprintf(stderr, "Corrupt data file: Requested buffer size %u exceeds max. buffer size.\n", in_flow_header.size);
-				break;
-			}
-			// make it at least the requested size
-			buffer_size = in_flow_header.size;
-			tmp = realloc((void *)in_buff, buffer_size);
-			if ( !tmp ) {
-				fprintf(stderr, "Can't reallocate buffer to %u bytes: %s\n", buffer_size, strerror(errno));
-				break;
-			}
-			in_buff = (common_record_t *)tmp;
-		}
-
-#ifdef COMPAT14
-		if ( Format14 ) 
-			ret = Compat14_ReadRecords(rfd, in_buff, &in_flow_header);
-		else
-			ret = read(rfd, in_buff, in_flow_header.size);
-#else
-		ret = read(rfd, in_buff, in_flow_header.size);
-#endif
-		if ( ret == 0 ) {
-			done = 1;
-			break;
-		} else if ( ret == -1 ) {
-			perror("Error reading data");
-			close(rfd);
-			if ( write_file ) {
-				/* Write stat info and close file */
-				CloseUpdateFile(wfd, &stat_record, file_blocks, GetIdent(), &string );
-				if ( string != NULL )
-					fprintf(stderr, "%s\n", string);
-			} 
-			break;
-		}
-		if ( in_flow_header.size != ret ) {
-			// Ups - this was a short read - most likely reading from the stdin pipe
-			// loop until we have requested size
-			size_t	request_size, total_size;
-			void *read_ptr;
-
-			total_size 	 = ret;
-			request_size = in_flow_header.size - total_size;
-			read_ptr 	 = (void *)((pointer_addr_t)in_buff + total_size);
-			do {
-				ret = read(rfd, read_ptr, request_size);
-				if ( ret == 0 ) {
-					break;
-				} else if ( ret < 0 ) {
-					perror("Error reading data");
-					break;
-				}
-				total_size += ret;
-				if ( total_size < in_flow_header.size ) {
-					request_size = in_flow_header.size - ret;
-					read_ptr 	 = (void *)((pointer_addr_t)in_buff + total_size);
-					request_size = in_flow_header.size - total_size;
-				}
-			} while ( ret > 0 && ( total_size < in_flow_header.size ));
-			
-			if ( total_size != in_flow_header.size ) {
-				// still unsuccessful
-#ifdef HAVE_SIZE_T_Z_FORMAT
-				fprintf(stderr, "Short read for netflow records: Expected %i, got %zu bytes!\n",
-					in_flow_header.size, total_size );
-#else
-				fprintf(stderr, "Short read for netflow records: Expected %i, got %lu bytes!\n",
-					in_flow_header.size, (unsigned long)total_size );
-#endif
+					// rfd == EMPTY_LIST
+					done = 1;
+				} // else continue with next file
 				continue;
-			} else {
-				// continue
-				ret = in_flow_header.size;
-			}
+	
+				break; // not really needed
+			default:
+				// successfully read block
+				total_bytes += ret;
+		}
+
+		if ( in_block_header.id != DATA_BLOCK_TYPE_1 ) {
+			fprintf(stderr, "Can't process block type %u. Skip block.\n", in_block_header.id);
+			skipped_records++;
+			continue;
 		}
-		total_bytes += ret;
 
 		flow_record = in_buff;
-		for ( i=0; i < NumReadRecords; i++ ) {
+		for ( i=0; i < in_block_header.NumBlocks; i++ ) {
 			total_flows++;
 			ExpandRecord( flow_record, &master_record);
 
@@ -702,7 +670,7 @@ extern int	Format14;
 
 
 			if ( wfd != 0 ) {
-				if ( (out_flow_header->size + flow_record->size) < OUTPUT_BUFF_SIZE && !zero_flows) {
+				if ( (out_block_header->size + flow_record->size) < BUFFSIZE ) {
 					memcpy(writeto, (void *)flow_record, flow_record->size);
 
 					if ( anon ) {
@@ -724,55 +692,43 @@ extern int	Format14;
 						}
 					} 
 	
-					out_flow_header->NumBlocks++;
-					out_flow_header->size += flow_record->size;
+					out_block_header->NumBlocks++;
+					out_block_header->size += flow_record->size;
 					writeto = (void *)((pointer_addr_t)writeto + flow_record->size);
 
 					// flush current buffer to disc
-					if ( out_flow_header->size > OUTPUT_FLUSH_LIMIT ) {
-						if ( write(wfd, out_buff, sizeof(data_block_header_t) + out_flow_header->size) <= 0 ) {
+					if ( out_block_header->size > OUTPUT_FLUSH_LIMIT ) {
+						if ( WriteBlock(wfd, out_block_header, compress) <= 0 ) {
 							fprintf(stderr, "Failed to write output buffer to disk: '%s'" , strerror(errno));
 						} else {
-							out_flow_header->size 		= 0;
-							out_flow_header->NumBlocks 	= 0;
-							writeto = (void *)((pointer_addr_t)out_flow_header + sizeof(data_block_header_t) );
+							out_block_header->size 		= 0;
+							out_block_header->NumBlocks 	= 0;
+							writeto = (void *)((pointer_addr_t)out_block_header + sizeof(data_block_header_t) );
 							file_blocks++;
 						}
 					}
 				}
 
 			} else if ( do_stat ) {
-					// we need to add this record to the stat hash
-					if ( has_aggregate_mask ) {
-						if ( (flow_record->flags & FLAG_IPV6_ADDR ) == 0 ) { // only v4 at the moment
-							uint64_t	*array64 = (uint64_t *)&master_record;
-							// mask ports
-							array64[3] &= AggregateMasks[0];
-							// mask IP addresses
-							array64[5] &= AggregateMasks[1];
-							array64[6] &= AggregateMasks[2];
-							array64[7] &= AggregateMasks[3];
-							array64[8] &= AggregateMasks[4];
-						}
-					}
-					AddStat(&in_flow_header, &master_record, flow_stat, element_stat);
+				// we need to add this record to the stat hash
+				AddStat(&in_block_header, &master_record, flow_stat, element_stat, has_aggregate_mask, AggregateMasks);
 			} else {
-					// if we need tp print out this record
-					if ( print_record ) {
-						print_record(&master_record, 1, &string, anon);
-						if ( string ) {
-							if ( limitflows ) {
-								if ( (stat_record.numflows <= limitflows) )
-									printf("%s\n", string);
-							} else 
+				// if we need tp print out this record
+				if ( print_record ) {
+					print_record(&master_record, 1, &string, anon, tag);
+					if ( string ) {
+						if ( limitflows ) {
+							if ( (stat_record.numflows <= limitflows) )
 								printf("%s\n", string);
-						}
+						} else 
+							printf("%s\n", string);
 					}
+				}
 
-					// if we need to sort the flows first -> insert into hash table
-					// they get 
-					if ( sort_flows ) 
-						InsertFlow(&master_record);
+				// if we need to sort the flows first -> insert into hash table
+				// they get 
+				if ( sort_flows ) 
+					InsertFlow(&master_record);
 			}
 
 			// Advance pointer by number of bytes for netflow record
@@ -792,8 +748,8 @@ extern int	Format14;
 	// flush output file
 	if ( wfd ) {
 		// flush current buffer to disc
-		if ( out_flow_header->NumBlocks && !zero_flows) {
-			if ( write(wfd, out_buff, sizeof(data_block_header_t) + out_flow_header->size) <= 0 ) {
+		if ( out_block_header->NumBlocks ) {
+			if ( WriteBlock(wfd, out_block_header, compress) <= 0 ) {
 				fprintf(stderr, "Failed to write output buffer to disk: '%s'" , strerror(errno));
 			} else {
 				file_blocks++;
@@ -803,7 +759,7 @@ extern int	Format14;
 		/* Stat info */
 		if ( write_file ) {
 			/* Write stat info and close file */
-			CloseUpdateFile(wfd, &stat_record, file_blocks, GetIdent(), &string );
+			CloseUpdateFile(wfd, &stat_record, file_blocks, GetIdent(), compress, &string );
 			if ( string != NULL )
 				fprintf(stderr, "%s\n", string);
 		} // else stdout
@@ -820,16 +776,18 @@ struct stat stat_buff;
 stat_record_t	sum_stat, *sr;
 printer_t 	print_header, print_record;
 nfprof_t 	profile_data;
-char 		c, *rfile, *Rfile, *Mdirs, *wfile, *ffile, *filter, *tstring, *stat_type;
+char 		*rfile, *Rfile, *Mdirs, *wfile, *ffile, *filter, *tstring, *stat_type;
 char		*byte_limit_string, *packet_limit_string, *print_mode, *record_header;
-char		*order_by, CryptoPAnKey[32];
-int 		ffd, ret, element_stat, fdump;
+char		*order_by, *query_file, *UnCompress_file, *nameserver, CryptoPAnKey[32];
+int 		c, ffd, ret, element_stat, fdump;
 int 		i, user_format, quiet, flow_stat, topN, aggregate, aggregate_mask;
-int 		print_stat, syntax_only, date_sorted, do_anonymize, zero_flows;
+int 		print_stat, syntax_only, date_sorted, do_anonymize, do_tag, compress;
+int			plain_numbers, pipe_output;
 time_t 		t_start, t_end;
 uint16_t	Aggregate_Bits;
 uint32_t	limitflows;
 uint64_t	AggregateMasks[AGGR_SIZE];
+char 		Ident[IdentLen];
 
 	rfile = Rfile = Mdirs = wfile = ffile = filter = tstring = stat_type = NULL;
 	byte_limit_string = packet_limit_string = NULL;
@@ -845,23 +803,31 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 	date_sorted		= 0;
 	total_bytes		= 0;
 	total_flows		= 0;
-	skipped_flows	= 0;
+	skipped_records	= 0;
 	do_anonymize	= 0;
+	do_tag			= 0;
 	quiet			= 0;
 	user_format		= 0;
-	zero_flows		= 0;
+	compress		= 0;
+	plain_numbers   = 0;
+	pipe_output		= 0;
+	nameserver		= NULL;
 
 	print_mode      = NULL;
 	print_header 	= NULL;
 	print_record  	= NULL;
+	query_file		= NULL;
+	UnCompress_file	= NULL;
 	record_header 	= "";
 	Aggregate_Bits	= 0xFFFF;	// set all bits
 
+	Ident[0] = '\0';
+
 	SetStat_DefaultOrder("flows");
 
 	for ( i=0; i<AGGR_SIZE; AggregateMasks[i++] = 0 ) ;
 
-	while ((c = getopt(argc, argv, "6aA:c:Ss:hn:f:qzr:w:K:M:ImO:R:XZt:Vv:l:L:o:")) != EOF) {
+	while ((c = getopt(argc, argv, "6aA:c:D:Ss:hn:i:j:f:qzr:v:w:K:M:NImO:R:XZt:TVv:l:L:o:")) != EOF) {
 		switch (c) {
 			case 'h':
 				usage(argv[0]);
@@ -877,6 +843,12 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 				}
 				aggregate_mask = 1;
 				break;
+			case 'D':
+				nameserver = optarg;
+				if ( !set_nameserver(nameserver) ) {
+					exit(255);
+				}
+				break;
 			case 'X':
 				fdump = 1;
 				break;
@@ -887,7 +859,7 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 				quiet = 1;
 				break;
 			case 'z':
-				zero_flows = 1;
+				compress = 1;
 				break;
 			case 'c':	
 				limitflows = atoi(optarg);
@@ -919,6 +891,9 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 			case 'L':
 				byte_limit_string = optarg;
 				break;
+			case 'N':
+				plain_numbers = 1;
+				break;
 			case 'f':
 				ffile = optarg;
 				break;
@@ -952,9 +927,6 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 			case 'R':
 				Rfile = optarg;
 				break;
-			case 'v':
-				fprintf(stderr, "Option no longer supported.\n");
-				break;
 			case 'w':
 				wfile = optarg;
 				break;
@@ -967,11 +939,29 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 				break;
 			case 'S':	// Compatibility with pre 1.4 -S option
 				stat_type = "record/packets/bytes";
+				fprintf(stderr, "WARNING: -S depricated! use -s record/packets/bytes instead. Option will get removed.\n");
                 if ( !SetStat(stat_type, &element_stat, &flow_stat) ) {
                     // Should never happen
                     exit(255);
                 } 
 				break;
+			case 'T':
+				do_tag = 1;
+				break;
+			case 'i':
+				strncpy(Ident, optarg, IDENT_SIZE);
+				Ident[IDENT_SIZE - 1] = 0;
+				if ( strchr(Ident, ' ') ) {
+					fprintf(stderr,"Ident must not contain spaces\n");
+					exit(255);
+				}
+				break;
+			case 'j':
+				UnCompress_file = optarg;
+				break;
+			case 'v':
+				query_file = optarg;
+				break;
 			case '6':	// print long IPv6 addr
 				Setv6Mode(1);
 				break;
@@ -988,10 +978,32 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 		filter = argv[optind];
 	}
 	
+	// Change Ident only
+	if ( rfile && strlen(Ident) > 0 ) {
+		char *err;
+		ChangeIdent(rfile, Ident, &err);
+		exit(0);
+	}
+
+	if ( query_file ) {
+		QueryFile(query_file);
+		exit(0);
+	}
+
+	if ( UnCompress_file ) {
+		UnCompressFile(UnCompress_file);
+		exit(0);
+	}
+
+	if ( (element_stat && !flow_stat) && aggregate_mask ) {
+		fprintf(stderr, "Warning: Aggregation ignored for element statistics\n");
+		aggregate_mask = 0;
+	}
+
 	if ( aggregate_mask && stat_type ) {
 		char *stat_conflict = VerifyStat(Aggregate_Bits);
 		if ( stat_conflict ) {
-			fprintf(stderr, "Selected aggregation masks stat '%s'!\n", stat_conflict);
+			fprintf(stderr, "Selected aggregation masks record field '%s'!\n", stat_conflict);
 			exit(255);
 		}
 	}
@@ -1072,6 +1084,10 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 					record_header = format_special_header();
 					user_format	  = 1;
 				} else {
+					// To support the pipe output format for element stats - check for pipe, and remember this
+					if ( strncasecmp(print_mode, "pipe", MAXMODELEN) == 0 ) {
+						pipe_output = 1;
+					}
 					// predefined static format
 					print_record  = printmap[i].func;
 					record_header = printmap[i].HeaderLine;
@@ -1124,7 +1140,7 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 	}
 
 	// if no filter is given, set the default ip filter which passes through every flow
-	if ( !filter ) 
+	if ( !filter  || strlen(filter) == 0 ) 
 		filter = "any";
 
 	Engine = CompileFilter(filter);
@@ -1177,34 +1193,34 @@ uint64_t	AggregateMasks[AGGR_SIZE];
 	nfprof_start(&profile_data);
 	sum_stat = process_data(wfile, element_stat, aggregate || flow_stat, date_sorted,
 						print_header, print_record, t_start, t_end, 
-						limitflows, aggregate_mask ? AggregateMasks : NULL, do_anonymize, zero_flows);
+						limitflows, aggregate_mask ? AggregateMasks : NULL, do_anonymize, do_tag, compress);
 	nfprof_end(&profile_data, total_flows);
 
 	if ( total_bytes == 0 )
 		exit(0);
 
 	if (aggregate) {
-		ReportAggregated(print_record, limitflows, date_sorted, do_anonymize);
+		ReportAggregated(print_record, limitflows, date_sorted, do_anonymize, do_tag);
 		Dispose_Tables(1, 0); // Free FlowTable
 	}
 
 	if (flow_stat || element_stat) {
-		ReportStat(record_header, print_record, topN, flow_stat, element_stat, do_anonymize);
+		ReportStat(record_header, print_record, topN, flow_stat, element_stat, do_anonymize, do_tag, pipe_output);
 		Dispose_Tables(flow_stat, element_stat);
 	} 
 
 	if ( date_sorted && !(aggregate || flow_stat || element_stat) ) {
-		PrintSortedFlows(print_record, limitflows, do_anonymize);
+		PrintSortedFlows(print_record, limitflows, do_anonymize, do_tag);
 		Dispose_Tables(1, 0);	// Free FlowTable
 	}
 
 	if ( !wfile && !quiet ) {
 		if (do_anonymize)
 			printf("IP addresses anonymized\n");
-		PrintSummary(&sum_stat);
+		PrintSummary(&sum_stat, plain_numbers);
  		printf("Time window: %s\n", TimeString(t_first_flow, t_last_flow));
-		printf("Total flows processed: %u, skipped: %u, Bytes read: %llu\n", 
-			total_flows, skipped_flows, total_bytes);
+		printf("Total flows processed: %u, Records skipped: %u, Bytes read: %llu\n", 
+			total_flows, skipped_records, (unsigned long long)total_bytes);
 		nfprof_print(&profile_data, stdout);
 	}
 	return 0;
