@@ -30,27 +30,33 @@
  *  
  *  $Author: peter $
  *
- *  $Id: nftest.c 70 2006-05-17 08:38:01Z peter $
+ *  $Id: nftest.c 97 2008-02-21 09:50:02Z peter $
  *
- *  $LastChangedRevision: 70 $
+ *  $LastChangedRevision: 97 $
  *	
  */
 
+#include "config.h"
+
 #include <stdio.h>
+#include <stdarg.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/time.h>
+#include <sys/param.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
 #include <string.h>
 #include <stdlib.h>
-
-#include "config.h"
+#include <unistd.h>
+#include <errno.h>
 
 #ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif
 
+#include "rbtree.h"
 #include "nfdump.h"
 #include "nftree.h"
 #include "nffile.h"
@@ -61,16 +67,35 @@
 extern char 	*CurrentIdent;
 
 FilterEngine_data_t	*Engine;
-uint32_t			byte_limit, packet_limit;
-int 				byte_mode, packet_mode;
 
+#define ICMP 1
 #define TCP	6
 #define UDP 17
 
+/* exported fuctions */
+void LogError(char *format, ...);
+
 int check_filter_block(char *filter, master_record_t *flow_record, int expect);
 
 void check_offset(char *text, pointer_addr_t offset, pointer_addr_t expect);
 
+void CheckCompression(char *filename);
+
+/* 
+ * some modules are needed for daemon code as well as normal stdio code 
+ * therefore a generic LogError is defined, which maps in this case
+ * to stderr
+ */
+void LogError(char *format, ...) {
+va_list var_args;
+
+	va_start(var_args, format);
+	vfprintf(stderr, format, var_args);
+	va_end(var_args);
+
+} // End of LogError
+
+
 int check_filter_block(char *filter, master_record_t *flow_record, int expect) {
 int ret, i;
 uint64_t	*block = (uint64_t *)flow_record;
@@ -90,7 +115,7 @@ uint64_t	*block = (uint64_t *)flow_record;
 		printf("Expected: %i, Found: %i\n", expect, ret);
 		printf("Record:\n");
 		for(i=0; i<=10; i++) {
-			printf("%3i %.16llx\n", i, block[i]);
+			printf("%3i %.16llx\n", i, (long long)block[i]);
 		}
 		if ( Engine->IdentList ) {
 			printf("Current Ident: %s, Ident 0 %s\n", CurrentIdent, Engine->IdentList[0]);
@@ -103,14 +128,100 @@ uint64_t	*block = (uint64_t *)flow_record;
 void check_offset(char *text, pointer_addr_t offset, pointer_addr_t expect) {
 
 	if ( offset == expect ) {
-		printf("Success: %s: %u\n", text, expect);
+		printf("Success: %s: %llu\n", text, (unsigned long long)expect);
 	} else {
-		printf("**** FAILED **** %s expected %u, evaluated %u\n", text, expect, offset);
+		printf("**** FAILED **** %s expected %llu, evaluated %llu\n", 
+			text, (unsigned long long)expect, (unsigned long long)offset);
 		// useless to continue
 		exit(255);
 	}
 }
 
+void CheckCompression(char *filename) {
+int i, rfd, wfd, compress;
+ssize_t	ret;
+stat_record_t *stat_ptr;
+data_block_header_t *block_header;
+char	*string;
+char outfile[MAXPATHLEN];
+void	*buff, *buff_ptr;
+
+struct timeval  	tstart[2];
+struct timeval  	tend[2];
+u_long usec, sec;
+double wall[2];
+
+	rfd = OpenFile(filename, &stat_ptr, &string);
+	if ( rfd < 0 ) {
+		fprintf(stderr, "%s\n", string);
+		return;
+	}
+	
+	// tmp filename for new output file
+	snprintf(outfile, MAXPATHLEN, "%s-tmp", filename);
+	outfile[MAXPATHLEN-1] = '\0';
+
+	buff = malloc(BUFFSIZE);
+	if ( !buff ) {
+		fprintf(stderr, "Buffer allocation error: %s", strerror(errno));
+		close(rfd);
+		return;
+	}
+	block_header = (data_block_header_t *)buff;
+	buff_ptr	 = (void *)((pointer_addr_t)buff + sizeof(data_block_header_t));
+
+	ret = ReadBlock(rfd, block_header, buff_ptr, &string);
+	if ( ret < 0 ) {
+		fprintf(stderr, "Error while reading data block. Abort.\n");
+		close(rfd);
+		unlink(outfile);
+		return;
+	}
+	close(rfd);
+
+	for ( compress=0; compress<=1; compress++ ) {
+		wfd = OpenNewFile(outfile, &string, 1);
+		if ( wfd < 0 ) {
+        	fprintf(stderr, "%s\n", string);
+        	return;
+    	}
+
+		gettimeofday(&(tstart[compress]), (struct timezone*)NULL);
+		for ( i=0; i<100; i++ ) {
+			if ( (ret = WriteBlock(wfd, block_header, compress)) <= 0 ) {
+				fprintf(stderr, "Failed to write output buffer to disk: '%s'" , strerror(errno));
+				close(rfd);
+				close(wfd);
+				unlink(outfile);
+				return;
+			}
+		}
+		gettimeofday(&(tend[compress]), (struct timezone*)NULL);
+
+		CloseUpdateFile(wfd, stat_ptr, 0, "none", compress, &string );
+		unlink(outfile);
+
+		if (tend[compress].tv_usec < tstart[compress].tv_usec) 
+			tend[compress].tv_usec += 1000000, --tend[compress].tv_sec;
+
+		usec = tend[compress].tv_usec - tstart[compress].tv_usec;
+		sec  = tend[compress].tv_sec - tstart[compress].tv_sec;
+
+		wall[compress] = (double)sec + ((double)usec)/1000000;
+	}
+
+	printf("100 write cycles, with size %u bytes\n", block_header->size);
+	printf("Uncompressed write time: %-.6fs size: %u\n", wall[0], block_header->size);
+	printf("Compressed write time  : %-.6fs size: %d\n", wall[1], (int32_t)ret);
+	printf("Ratio                  : 1:%-.3f\n", (double)ret/(double)block_header->size);
+
+	if ( wall[0] < wall[1] )
+		printf("You should run nfcapd without compression\n");
+	else
+		printf("You can run nfcapd with compression (-z)\n");
+	
+} // End of CheckCompression
+
 int main(int argc, char **argv) {
 master_record_t flow_record;
 uint64_t *blocks, l;
@@ -137,6 +248,13 @@ value64_t	v;
         printf("**** FAILED **** val32/64 union check failed!\n" );
 		exit(255);
 	}
+
+	if ( argc == 2 ) {
+		CheckCompression(argv[1]);
+		exit(0);
+	}
+
+
 	size = sizeof(common_record_t) - sizeof(uint8_t[4]);
 	memset((void *)&flow_record, 0, sizeof(master_record_t));
 	blocks = (uint64_t *)&flow_record;
@@ -153,15 +271,17 @@ value64_t	v;
 	check_offset("tos      Offset", (unsigned int)((pointer_addr_t)&flow_record.tos     - (pointer_addr_t)&blocks[OffsetTos]), 7);
 
 #ifdef HAVE_SIZE_T_Z_FORMAT
-	printf("Pointer Size : %zu\n", sizeof(blocks));
-	printf("Time_t  Size : %zu\n", sizeof(now));
-	printf("int     Size : %zu\n", sizeof(int));
-	printf("long    Size : %zu\n", sizeof(long));
+	printf("Pointer  Size : %zu\n", sizeof(blocks));
+	printf("Time_t   Size : %zu\n", sizeof(now));
+	printf("int      Size : %zu\n", sizeof(int));
+	printf("long     Size : %zu\n", sizeof(long));
+	printf("longlong Size : %zu\n", sizeof(long long));
 #else
-	printf("Pointer Size : %lu\n", (unsigned long)sizeof(blocks));
-	printf("Time_t  Size : %lu\n", (unsigned long)sizeof(now));
-	printf("int     Size : %lu\n", (unsigned long)sizeof(int));
-	printf("long    Size : %lu\n", (unsigned long)sizeof(long));
+	printf("Pointer  Size : %lu\n", (unsigned long)sizeof(blocks));
+	printf("Time_t   Size : %lu\n", (unsigned long)sizeof(now));
+	printf("int      Size : %lu\n", (unsigned long)sizeof(int));
+	printf("long     Size : %lu\n", (unsigned long)sizeof(long));
+	printf("longlong Size : %lu\n", (unsigned long)sizeof(long long));
 #endif
 
 	flow_record.flags	 = 0;
@@ -180,7 +300,6 @@ value64_t	v;
 	flow_record.prot	 = TCP;
 	ret = check_filter_block("any", &flow_record, 1);
 	ret = check_filter_block("not any", &flow_record, 0);
-	ret = check_filter_block("tcp", &flow_record, 1);
 	ret = check_filter_block("proto tcp", &flow_record, 1);
 	ret = check_filter_block("proto udp", &flow_record, 0);
 	flow_record.prot = UDP;
@@ -198,6 +317,15 @@ value64_t	v;
 	ret = check_filter_block("proto 47", &flow_record, 1);
 	ret = check_filter_block("proto 42", &flow_record, 0);
 
+	flow_record.prot = 1;
+	flow_record.dstport = 250; // -> icmp code 250
+	ret = check_filter_block("icmp-code 250", &flow_record, 1);
+	ret = check_filter_block("icmp-code 251", &flow_record, 0);
+	flow_record.dstport = 3 << 8; // -> icmp type 8
+	ret = check_filter_block("icmp-type 3", &flow_record, 1);
+	ret = check_filter_block("icmp-type 4", &flow_record, 0);
+
+
 	inet_pton(PF_INET6, "fe80::2110:abcd:1234:5678", flow_record.v6.srcaddr);
 	inet_pton(PF_INET6, "fe80::1104:fedc:4321:8765", flow_record.v6.dstaddr);
 	flow_record.v6.srcaddr[0] = ntohll(flow_record.v6.srcaddr[0]);
@@ -217,6 +345,9 @@ value64_t	v;
 	ret = check_filter_block("not ip fe80::2110:abcd:1234:5678", &flow_record, 0);
 	ret = check_filter_block("not ip fe80::2110:abcd:1234:5679", &flow_record, 1);
 
+	ret = check_filter_block("src ip in [fe80::2110:abcd:1234:5678]", &flow_record, 1);
+	ret = check_filter_block("src ip in [fe80::2110:abcd:1234:5679]", &flow_record, 0);
+
 	inet_pton(PF_INET6, "fe80::2110:abcd:1234:0", flow_record.v6.srcaddr);
 	flow_record.v6.srcaddr[0] = ntohll(flow_record.v6.srcaddr[0]);
 	flow_record.v6.srcaddr[1] = ntohll(flow_record.v6.srcaddr[1]);
@@ -288,6 +419,17 @@ value64_t	v;
 	ret = check_filter_block("not host 172.32.7.16", &flow_record, 0);
 	ret = check_filter_block("not host 172.32.7.17", &flow_record, 1);
 
+	ret = check_filter_block("src ip in [172.32.7.16]", &flow_record, 1);
+	ret = check_filter_block("src ip in [172.32.7.17]", &flow_record, 0);
+	ret = check_filter_block("src ip in [10.10.10.11]", &flow_record, 0);
+	ret = check_filter_block("dst ip in [10.10.10.11]", &flow_record, 1);
+	ret = check_filter_block("ip in [10.10.10.11]", &flow_record, 1);
+	ret = check_filter_block("ip in [172.32.7.16]", &flow_record, 1);
+	ret = check_filter_block("src ip in [172.32.7.16 172.32.7.17 10.10.10.11 10.10.10.12 ]", &flow_record, 1);
+	ret = check_filter_block("dst ip in [172.32.7.16 172.32.7.17 10.10.10.11 10.10.10.12 ]", &flow_record, 1);
+	ret = check_filter_block("ip in [172.32.7.16 172.32.7.17 10.10.10.11 10.10.10.12 ]", &flow_record, 1);
+	ret = check_filter_block("ip in [172.32.7.17 172.32.7.18 10.10.10.12 10.10.10.13 ]", &flow_record, 0);
+
 	flow_record.srcport = 63;
 	flow_record.dstport = 255;
 	ret = check_filter_block("src port 63", &flow_record, 1);
@@ -319,6 +461,17 @@ value64_t	v;
 	ret = check_filter_block("dst port lt 256", &flow_record, 1);
 	ret = check_filter_block("dst port < 255", &flow_record, 0);
 
+	ret = check_filter_block("src port in [ 62 63 64 ]", &flow_record, 1);
+	ret = check_filter_block("src port in [ 62 64 65 ]", &flow_record, 0);
+	ret = check_filter_block("dst port in [ 254 255 256 ]", &flow_record, 1);
+	ret = check_filter_block("dst port in [ 254 256 257 ]", &flow_record, 0);
+	ret = check_filter_block("port in [ 62 63 64 ]", &flow_record, 1);
+	ret = check_filter_block("port in [ 254 255 256 ]", &flow_record, 1);
+	ret = check_filter_block("port in [ 62 63 64 254 255 256 ]", &flow_record, 1);
+	ret = check_filter_block("port in [ 62 63 64 254 256 ]", &flow_record, 1);
+	ret = check_filter_block("port in [ 62 64 254 256 ]", &flow_record, 0);
+	ret = check_filter_block("not port in [ 62 64 254 256 ]", &flow_record, 1);
+
 	flow_record.srcas = 123;
 	flow_record.dstas = 456;
 	ret = check_filter_block("src as 123", &flow_record, 1);
@@ -330,6 +483,18 @@ value64_t	v;
 	ret = check_filter_block("as 124", &flow_record, 0);
 	ret = check_filter_block("as 457", &flow_record, 0);
 
+	ret = check_filter_block("src as in [ 122 123 124 ]", &flow_record, 1);
+	ret = check_filter_block("dst as in [ 122 124 125 ]", &flow_record, 0);
+	ret = check_filter_block("dst as in [ 455 456 457 ]", &flow_record, 1);
+	ret = check_filter_block("dst as in [ 455 457 458 ]", &flow_record, 0);
+	ret = check_filter_block("as in [ 122 123 124 ]", &flow_record, 1);
+	ret = check_filter_block("as in [ 455 456 457 ]", &flow_record, 1);
+	ret = check_filter_block("as in [ 122 123 124 455 456 457]", &flow_record, 1);
+	ret = check_filter_block("as in [ 122 123 124 455 457]", &flow_record, 1);
+	ret = check_filter_block("as in [ 122 124 455 456 457]", &flow_record, 1);
+	ret = check_filter_block("as in [ 122 124 455 457]", &flow_record, 0);
+	ret = check_filter_block("not as in [ 122 124 455 457]", &flow_record, 1);
+
 	ret = check_filter_block("src net 172.32/16", &flow_record, 1);
 	ret = check_filter_block("src net 172.32.7/24", &flow_record, 1);
 	ret = check_filter_block("src net 172.32.7.0/27", &flow_record, 1);
